use std::{
    fmt,
    sync::mpsc,
    thread,
    time::Duration,
};

use rayon::ThreadPool;
use thiserror::Error;

use crate::controller::*;

mod controller;

fn main() -> Result <(), TaskError>
{
    let pool = rayon::ThreadPoolBuilder::new ().build ().unwrap ();
    let mut driver = Driver::new (&pool);
    driver.main ()?;
    Ok (())
}

struct ThreadSystem <'a>
{
    pool: &'a ThreadPool,
    send: mpsc::SyncSender <TaskOutput>,
}

struct Driver <'a>
{
    thread_sys: ThreadSystem <'a>,
    recv: mpsc::Receiver <TaskOutput>,
    
    ctl: Controller,
    
    capture: Task <Capture, ()>,
    encoder: Task <Encoder, EncoderTaskMetadata>,
    transmitter: Task <Transmitter, ()>,
}

impl <'a> Driver <'_>
{
    fn new (pool: &'a ThreadPool) -> Driver <'a>
    {
        let (send, recv) = mpsc::sync_channel (8);
        
        let thread_sys = ThreadSystem
        {
            pool,
            send,
        };
        
        Driver
        {
            thread_sys,
            recv,
            
            ctl: Default::default (),
            
            capture: Capture::default ().into (),
            encoder: Encoder::default ().into (),
            transmitter: Transmitter::default ().into (),
        }
    }
    
    fn main (&mut self) -> Result <(), TaskError>
    {
        for _ in 0..10
        {
            let ev = self.ctl.poll ();
            
            if let Some (event) = ev.tx
            {
                dbg! (&event);
                self.handle_tx_event (event)?;
                
                match self.recv.try_recv ()
                {
                    Ok (output) => self.handle_task_output (output)?,
                    Err (_) => (),
                }
            }
            else
            {
                match self.recv.recv ()
                {
                    Ok (output) => self.handle_task_output (output)?,
                    Err (_) => (),
                }
            }
        }
        
        Ok (())
    }
    
    fn handle_task_output (&mut self, output: TaskOutput) -> Result <(), TaskError>
    {
        dbg! (&output);
        match output
        {
            TaskOutput::TxCapture ((cap, buf_raw)) =>
            {
                self.capture.stop (cap)?;
                self.ctl.handle_capture_frame (buf_raw).unwrap ();
            },
            TaskOutput::TxEncode (enc) =>
            {
                let _metadata = self.encoder.stop (enc)?;
                self.ctl.handle_encoder_finished ();
            },
            TaskOutput::TxTransmit ((tx, busy)) =>
            {
                self.transmitter.stop (tx)?;
                self.ctl.handle_network_busy (busy);
            }
        }
        
        Ok (())
    }
    
    fn handle_tx_event (&mut self, event: TxPipelineEvent) -> Result <(), TaskError>
    {
        let dur_capture = Duration::from_millis (30);
        let dur_encode = Duration::from_millis (20);
        let dur_transmit = Duration::from_millis (200);
        
        let pool = &self.thread_sys.pool;
        let send = self.thread_sys.send.clone ();
        
        match event
        {
            TxPipelineEvent::Capture => {
                let cap = self.capture.start (())?;
                
                pool.spawn (move || 
                {
                    thread::sleep (dur_capture);
                    let buf_raw = BufRaw
                    {
                        buf: vec![],
                    };
                    send.send (TaskOutput::TxCapture ((cap, buf_raw))).unwrap ();
                });
            },
            TxPipelineEvent::PollEncoder =>
            {
                let encoder = self.encoder.try_inner_mut ()?;
                
                let mut buf = vec! [];
                
                let may_have_data = encoder.poll_encoded (&mut buf);
                if may_have_data
                {
                    self.ctl.handle_encoded_packet (Some (BufEncoded {buf})).unwrap ();
                }
                else
                {
                    self.ctl.handle_encoded_packet (None).unwrap ();
                }
            }
            TxPipelineEvent::Encode (buf_raw) => 
            {
                let mut encoder = self.encoder.start (EncoderTaskMetadata {})?;
                
                pool.spawn (move || 
                {
                    thread::sleep (dur_encode);
                    encoder.try_handle_insert_data (&buf_raw.buf).unwrap ();
                    send.send (TaskOutput::TxEncode (encoder)).unwrap ();
                });
            },
            TxPipelineEvent::Transmit (_buf_enc) => {
                let tx = self.transmitter.start (())?;
                pool.spawn (move ||
                {
                    thread::sleep (dur_transmit);
                    send.send (TaskOutput::TxTransmit ((tx, false))).unwrap ();
                });
            },
        }
        
        Ok (())
    }
}

/// This is probably just a fancy Cell or something.

enum Task <S, R>
{
    Stopped (S),
    Running (R),
}

#[derive (Debug, Error)]
enum TaskError
{
    #[error ("tried to start already-running task")]
    AlreadyRunning,
    #[error ("tried to stop already-stopped task")]
    AlreadyStopped,
    #[error ("cannot access a task's inner object while it's running")]
    CantAccessWhileRunning,
}

impl <S, R> From <S> for Task <S, R>
{
    fn from (x: S) -> Self
    {
        Self::Stopped (x)
    }
}

impl <S, R> Task <S, R>
{
    fn try_inner (&self) -> Result <&S, TaskError>
    {
        match self
        {
            Self::Running (_) => Err (TaskError::CantAccessWhileRunning),
            Self::Stopped (x) => Ok (x),
        }
    }
    
    fn try_inner_mut (&mut self) -> Result <&mut S, TaskError>
    {
        match self
        {
            Self::Running (_) => Err (TaskError::CantAccessWhileRunning),
            Self::Stopped (x) => Ok (x),
        }
    }
    
    fn start (&mut self, x: R) -> Result <S, TaskError>
    {
        match std::mem::replace (self, Self::Running (x))
        {
            Self::Running (_) => Err (TaskError::AlreadyRunning),
            Self::Stopped (x) => Ok (x),
        }
    }
    
    fn stop (&mut self, x: S) -> Result <R, TaskError>
    {
        match std::mem::replace (self, Self::Stopped (x))
        {
            Self::Stopped (_) => Err (TaskError::AlreadyStopped),
            Self::Running (x) => Ok (x),
        }
    }
}

struct EncoderTaskMetadata
{
    
}

enum TaskOutput
{
    TxCapture ((Capture, BufRaw)),
    TxEncode (Encoder),
    TxTransmit ((Transmitter, bool)),
}

impl fmt::Debug for TaskOutput {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self
        {
            Self::TxCapture (_) => f.debug_tuple ("TaskOutput::TxCapture").finish (),
            Self::TxEncode (_) => f.debug_tuple ("TaskOutput::TxEncode").finish (),
            Self::TxTransmit (_) => f.debug_tuple ("TaskOutput::TxTransmit").finish (),
        }
    }
}

/// Wraps non-thread-safe capture device like a v4l webcam

#[derive (Default)]
struct Capture
{
    
}

impl Capture
{
    /// Blocks until the capture device gets us a frame
    
    fn wait_for_frame (&mut self, _buf: &mut [u8]) -> bool
    {
        false
    }
}

/// Wraps non-thread-safe encoder state like from ffmpeg

#[derive (Default)]
struct Encoder
{
    /// AVCodecContext
    ctx: (),
    
    /// Encoded AVPacket
    pkt_enc: Option <()>,
    
    /// Bogus debugging thing
    internal_buffer: usize,
}

impl Encoder
{
    /// If any encoded bytes are ready, and the caller's buffer 
    /// is big enough, copy encoded bytes into the caller's buffer, 
    /// consuming them from internal buffers.
    /// 
    /// Won't block
    
    fn poll_encoded (&mut self, _buf_enc: &mut [u8]) -> bool
    {
        if let Some (_pkt) = self.pkt_enc.take ()
        {
            // Do copy
            return true;
        }
        
        if self.internal_buffer >= 2
        {
            self.internal_buffer -= 2;
            // Do copy
            return true;
        }
        
        false
    }
    
    /// Insert raw data. If the internal buffers are full, the new
    /// data is rejected.
    /// Call poll_encoded before calling this, to drain internal
    /// buffers if needed.
    /// 
    /// Blocks if the internals perform encoding
    
    fn try_handle_insert_data (&mut self, _buf_raw: &[u8]) -> Result <(), ControlError>
    {
        if self.pkt_enc.is_some ()
        {
            return Err (ControlError::AlreadyHaveRawData);
        }
        
        if self.internal_buffer >= 3
        {
            return Err (ControlError::AlreadyHaveRawData);
        }
        
        self.internal_buffer += 3;
        
        Ok (())
    }
}

/// Wraps a network transmitter like a TCP send stream or a QUIC
/// outgoing uni stream

#[derive (Default)]
struct Transmitter
{
    
}

impl Transmitter
{
    /// Tries to transmit data over the network.
    /// If Transmitter sees congestion, it may reject the transmission
    /// until a later time.
    
    fn try_send (&mut self, _now: std::time::Instant, _buf: &[u8]) -> bool
    {
        false
    }
}
